# Method for calculating input values to PI methods
# based on AR(1) model with deterministic seasons
# and NIG distributed residuals
#
# Arguments:
# m       -   time steps until t=T
# data    -   historical values of S_T, ending at t=0
# 
setupPI = function(m,data)
{
  # estimate parameters from the complete data set
  # in order to obtain starting values for next LS procedure
  LStest = ordinaryLS(log(data))
  startval = LStest$estimate

  # fit a seasonality function to the m last historical log prices
  LSfit = ordinaryLS(log(tail(data,m)),init=startval,typsize=startval)

  # plot log price with estimated seasonality function
  #plot((1-m):0, log(tail(data,m)), type="l", ylab="Log price", xlab="Time (days)")
  #lines((1-m):0, g(LSfit$estimate, 1:m), col=2)

  # calculate residuals after deseasonalizing, and fit AR(1) model
  detrended = log(tail(data,m))-g(LSfit$estimate, 1:m)
  arfit = ar(detrended,order=1)
  print(paste("AR parameter",arfit$ar))
  residuals = removeAR(detrended, arfit)

  #plot detrended and ar residuals
  #dev.new()
  #plot(1:m, detrended, type="l")
  #lines(2:m, residuals, col=2)
  #print(mean(detrended))
  #print(mean(residuals))
  
  # estimate NIG parameters
  NIGpar = estimateNIGparams(residuals)
  p = ppoints(length(residuals))
  quant = qnig(p,NIGpar[1],NIGpar[2],NIGpar[3],NIGpar[4])
  #dev.new()
  #plot(quant, sort(residuals))
  #lines(quant,quant)

  # do k MC runs to find approximate grid size
  k = 1000
  MC = c()
  for(i in 1:k)
  {
    # simulate m NIG-residuals
    nigvar = rnig(m,NIGpar[1],NIGpar[2],NIGpar[3],NIGpar[4])

    # simulate Autoregressive process X
    X = c(tail(detrended,1))
    alpha = arfit$ar-1
    for(j in 1:m)
    {
      dX = alpha*tail(X,1) + nigvar[j]
      X = c(X,tail(X,1)+dX)
    }
    X = tail(X,m)
    result = X + logLambda(LSfit$estimate, (m+1):(2*m))
    MC = cbind(MC,result)
    #if((i%%100)==0) print(paste("Progress:",i))
  }

  # plot fist simulation with historical and seasonfunction
  #plot((1-m):0,log(tail(nordpool$price,m)),type="l",xlim=c(-m,m))
  #lines((1-m):m,g(LSfit$estimate,((1-m):m)+m), col=3,type="l")
  #lines(1:m,MC[,1])

  # find range around the seasonal function
  detrendMC = MC - logLambda(LSfit$estimate, (m+1):(2*m))
  mcrange = c()
  for (i in 1:m)
  {
    mcrange = rbind(mcrange, range(detrendMC[i,]))
  }

  # setup PDF grid (expand range by 100% to be sure)
  gridmin = min(detrendMC) - 0.5*(max(detrendMC)-min(detrendMC))
  gridmax = max(detrendMC) + 0.5*(max(detrendMC)-min(detrendMC))

  # compile setup object
  setup = NULL
  setup$X0 = tail(detrended,1)
  setup$range = range(detrendMC)
  setup$gridrange = c(gridmin,gridmax)
  setup$steprange = mcrange
  setup$seasonpar = LSfit$estimate
  setup$arpar = arfit$ar
  setup$nigpar = NIGpar
  setup$m = m
  # return setup object
  setup
}



# Use the PI approach to approximate PDF of log(S_T)
#
#
# Arguments:
# m         -   time steps until t=T
# data      -   historical values of S_T, ending at t=0
# setup     -   output from the setupPI function
# N         -   number of gridpoints for representing PDF
# K         -   number of gridpoints in the integration grid
# zero      -   any density less than this is set to 0
#
PIdensity = function(m,data,setup=NULL, N=100, K=1000, zero=10^-6)
{
  ptm = proc.time()

  # we need K to be an odd number i.o. to use Simpsons method
  if(K%%2==0) K = K+1

  # create setup if not submitted as argument
  if(is.null(setup))
    setup = setupPI(m,data)

  # setup PDF grid (expand range by 100% to be sure)
  gridmin = setup$gridrange[1]
  gridmax = setup$gridrange[2]
  grid = seq(gridmin,gridmax,length.out=N)

  # setup fine grid for integration
  #fgrid = seq(gridmin,gridmax,length.out=K)
  #fgridspace = (gridmax-gridmin)/(K-1)

  # calculate and store values of propagator: p(x|x')
  # xMat contains x' values
  # propMat contains corresponding p(x|x')
  # x is given by grid[row index]
  propMat = c()
  tempMat = c()
  xMat = c()
  # test run to find location of probability mass
  for(i in 1:N)
  {
    Xval = grid[i]   # value of x for this col
    condpdf = dnig(grid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
    tempMat = cbind(tempMat, condpdf)
  }
  # find grid and propagator densities for each x
  for(i in 1:N)
  {
    # find limits and grid for this propagator row
    # to avoid calculations where propagator is zero
    limits = range(grid[which(tempMat[i,]>zero)])
    xMat = rbind(xMat,seq(limits[1],limits[2],length.out=K))

    # find densities for this propagator row
    row = c()
    for(j in 1:K)
    {
      row = c(row, dnig(grid[i], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*xMat[i,j])))
    }
    propMat = rbind(propMat,row)
  }

  # find distribution at t=1 (do calculations in deseasonalized space)
  X0 = log(tail(data,1))-logLambda(setup$seasonpar, m)
  pdf = c(0,dnig(grid[2:(N-1)], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4] + setup$arpar*X0)),0)
  pdfspline = splinefun(grid, pdf,method="natural")
  #plot(fgrid, pdfspline(fgrid),type="l",xlim=c(-0.6,0.5),main="Densities propagated with Simpson's Rule", xlab=expression(x[t]), ylab=expression(p(x[t])))

  print("starting iteration forward in time")
  # iterate through time steps
  for(t in 2:m)
  {
    # iterate through grid points
    newpdf = c(0)
    for(i in 2:(N-1))
    {
      # perform integration
      dens = 0
      h = diff(xMat[i,1:2])
      integrand = propMat[i,] * pdfspline(xMat[i,])
      for(j in seq(1,K-2,2))
      {
        simpson = (h/3) * (integrand[j] + 4*integrand[j+1] + integrand[j+2])
        dens = dens + simpson
      }
      newpdf = c(newpdf,dens)
    }
    newpdf = c(newpdf,0)
    pdfspline = splinefun(grid,newpdf,method="natural")

    if(t%%30==0) print(paste("Progress:",t))
#    if(t%%5==0 & t<=30)
#    {
#      c = (t/5)+1
#      print(c)
#      s = seq(gridmin,gridmax,0.01)
#      lines(s,pdfspline(s),col=c)
#    }
#    if(t>30)
#    {
#        legend("topright",legend=c("t=1","t=5","t=10","t=15","t=20","t=25","t=30"),col=c(1,2,3,4,5,6,7),lty=c(1,1,1,1,1,1,1), inset=0.05)
#        break
#    }
  }

  # find probability mass of spline
  mass = integrate(pdfspline,gridmin,gridmax)
  print(mass)



  #colnames(tempMat)=NULL
  #print(tempMat[50,])
  #print(range(which(tempMat[50,]>zero)))
  #debugrange = fgrid[range(which(tempMat[50,]>zero))]
  #print(debugrange)
  #print(xMat[50,])
  #print(dnig(fgrid[50], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*xMat[50,K])*dt))


  # plot p(x|x') asf x'
  #plot(xMat[50,],pdfMat[50,],type="l",xlim=range(fgrid))
  #points(fgrid[50],0)
  #lines(xMat[70,],pdfMat[70,])
  #points(fgrid[70],0)
  #lines(xMat[110,],pdfMat[110,])
  #points(fgrid[110],0)
  #dev.new()
  #plot(xMat[70,50:150],pdfMat[70,50:150],type="l")
  #points(fgrid[70],0)
  #print(pdfMat[50,])
  #print(pdfMat[100,0:600])
  #lines(xMat[200,],pdfMat[200,],type="l")
  #lines(xMat[800,],pdfMat[800,],type="l")
  #lines(xMat[400,],pdfMat[400,],type="l")
  #lines(xMat[600,],pdfMat[600,],type="l")
  #points(c(fgrid[100],fgrid[200],fgrid[400],fgrid[600]),c(0,0,0,0),pty="x")
  #dev.new()
  #plot(fgrid[100:250],pdfMat[200,100:250],type="l")
  
  result = NULL
  result$time = proc.time() - ptm
  result$setup = setup
  result$density = (function(x){pdfspline(x)/mass$val})
  result$range = setup$gridrange
  result
}

# Run Monte Carlo simulations to obtain density estimate
#
# Input:
# n       - number of MC runs
# m       - number of days in simulation
# setup   - object that contains model parameters
#
MonteCarloDensity = function(setup,n=10^6)
{
  ptm <- proc.time()
  mc = rep(0,n)
  for(i in 1:n)
  {
    # simulate m NIG-residuals
    nigvar = rnig(setup$m,setup$nigpar[1],setup$nigpar[2],setup$nigpar[3],setup$nigpar[4])

    # simulate Autoregressive process X
    X = setup$X0
    alpha = setup$arpar-1
    for(j in 1:setup$m)
    {
      dX = alpha*X + nigvar[j]
      X = X+dX
    }
    mc[i] = X
    if((i%%10000)==0)
    {
      print(paste("Progress:",i))
      timeleft = (proc.time()[3]-ptm[3])*(n-i) / (i*60)
      print(paste("Estimated time left:",timeleft,"minutes"))
    }
  }
  timer=proc.time()-ptm

  result = NULL
  result$n = n
  result$mc = mc
  result$time=timer
  result$setup = setup
  result
}


# Approximate probability density by using
# discrete probabilities and transition matrix
#
# Input:
# N       - number of points in grid
# setup   - object containing model parameters
#
#
transmatDensity = function(setup, N=100)
{
  ptm = proc.time()

  # setup PDF grid (expand range by 100% to be sure)
  gridmin = setup$gridrange[1]
  gridmax = setup$gridrange[2]
  grid = seq(gridmin,gridmax,length.out=N)

  # construct matrix of probability densities
  P = c()
  for(i in 1:N)
  {
    # value of X at grid point i
    Xval = grid[i]
    # get conditional densities for column i, normalize and add to matrix
    condpdf = dnig(grid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
    condpdf = condpdf/(sum(condpdf))
    P = cbind(P, condpdf)
  }
  
  # probability density vector at t=1
  p1 = dnig(grid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*setup$X0))
  g1 = p1/sum(p1)

  # propagate probability forward to t=m
  gm = powmat(P,setup$m-1) %*% g1
  pm = powmat(P,setup$m-1) %*% p1
  # set densities to zero at each end, and create spline
  pm[0]=pm[N]=0
  pspl = splinefun(grid,pm,method="natural")

  # find probability mass of pspl
  mass = integrate(pspl,gridmin,gridmax)

  # construct and return result object
  result = NULL
  result$time = proc.time()-ptm
  result$setup = setup
  result$grid = grid
  result$gm = gm
  result$density = (function(x){pspl(x)/mass$val})
  result$P = P
  result
}


# Find the n'th power of a matrix
#
# M   - the matrix
# n   - the power
powmat = function(M,n)
{
  result = diag(1, ncol(M))
  while(n>0)
  {
    if(n %% 2 != 0)
    {
      result = result %*% M
      n = n-1
      if(n==0) break
    }
    M = M %*% M
    n = n/2
  }
  result
}

# Find the n'th moment around the mean
#
# n - the order of the moment to be found
# dens - density function
# xlim - limits of integration (typically: setup$gridrange)
#
nthmoment = function(n, dens, xlim)
{
  mean = integrate(momentintegrand,xlim[1],xlim[2],mu=0,n=1,dens=dens)

  if(n>1)
  {
    result = integrate(momentintegrand, xlim[1],xlim[2],mu=mean$val,n=n,dens=dens)
  }
  else
    result = mean

  result$val
}

# Return integrand function for moment integrals
#
# dens - density
# n    - order of moment0
# mu   - mean of density function
# x    - input to density function
#
momentintegrand = function(x,mu,n,dens)
{
  (x-mu)^n*dens(x)
}

# Return skewness of density function
#
# dens  - density function
# xlim  - limits of integration (typically: setup$gridrange)
densSkew = function(dens,xlim)
{
  nthmoment(3,dens,xlim)/(nthmoment(2,dens,xlim)^(3/2))
}

# Return kurtosis of density function
#
# dens  - density function
# xlim  - limits of integration (typically: setup$gridrange)
densKurt = function(dens,xlim)
{
  nthmoment(4,dens,xlim)/(nthmoment(2,dens,xlim)^2) - 3
}

# Return log price density function
#
# Input:
# PIobj     - PI object
# (returned from transmatDens or PIdens)
#
logpriceDensity = function(PIobj)
{
  (function(lnS)
  {
    dens = rep(0,length(lnS))
    x = lnS - logLambda(PIobj$setup$seasonpar,2*PIobj$setup$m)
    notzero = which(x>PIobj$setup$gridrange[1] & x<PIobj$setup$gridrange[2])
    dens[notzero] = PIobj$density(x[notzero])
    dens
  })
}


# Return value of European call option
# calculated from density at expiry
#
# Input:
# PIobj - returned
# E     - exercise price
#
valueEuropeanCall = function(PIobj, E)
{
  # deterministic mean of log price
  logLambda = logLambda(PIobj$setup$seasonpar, 2*PIobj$setup$m)
  
  # limits of integration for log S density
  limits = PIobj$setup$gridrange + logLambda

  # check if exercise price is small enough
  if(log(E)>limits[2])
    value = 0
  else
  {
    # get density function for log S
    logdens = logpriceDensity(PIobj)

    # create integrand function
    integrand = (function(lnS){(exp(lnS)-E)*logdens(lnS)})

    # perform integration
    value = integrate(integrand,max(log(E),limits[1]), limits[2])$val
  }

  value
}

# Return value of European put option
# calculated from density at expiry
#
# Input:
# PIobj - returned
# E     - exercise price
#
valueEuropeanPut = function(PIobj, E)
{
  # deterministic mean of log price
  logLambda = logLambda(PIobj$setup$seasonpar, 2*PIobj$setup$m)
  
  # limits of integration for log S density
  limits = PIobj$setup$gridrange + logLambda

  # check if exercise price is small enough
  if(log(E)<limits[1])
    value = 0
  else
  {
    # get density function for log S
    logdens = logpriceDensity(PIobj)

    # create integrand function
    integrand = (function(lnS){(E-exp(lnS))*logdens(lnS)})

    # perform integration
    value = integrate(integrand,limits[1], min(log(E),limits[2]))$val
  }

  value
}

# Return value of forward contract
# calculated from density at maturity
#
# Input:
# PIobj - returned
# F     - forward price
#
valueForward = function(PIobj, F)
{
  # deterministic mean of log price
  logLambda = logLambda(PIobj$setup$seasonpar, 2*PIobj$setup$m)
  
  # limits of integration for log S density
  limits = PIobj$setup$gridrange + logLambda

  # get density function for log S
  logdens = logpriceDensity(PIobj)

  # create integrand function
  integrand = (function(lnS){(exp(lnS)-F)*logdens(lnS)})

  # perform integration
  value = integrate(integrand,limits[1], limits[2])$val

  value
}


#  Value European call option from MC results
#
#  Input:
#  MCobj  - result from MonteCarloDensity-function
#  E      - exercise price
#
mcEuropeanCall = function(MCobj,E)
{
  logLambda = logLambda(MCobj$setup$seasonpar, 2*MCobj$setup$m)
  values = exp(MCobj$mc + logLambda)-E
  values[which(values<0)]=0
  mean(values)
}

#  Value European put option from MC results
#
#  Input:
#  MCobj  - result from MonteCarloDensity-function
#  E      - exercise price
#
mcEuropeanPut = function(MCobj,E)
{
  logLambda = logLambda(MCobj$setup$seasonpar, 2*MCobj$setup$m)
  values = E-exp(MCobj$mc + logLambda)
  values[which(values<0)]=0
  mean(values)
}

#  Value forward contract from MC results
#
#  Input:
#  MCobj  - result from MonteCarloDensity-function
#  F      - forward price
#
mcForward = function(MCobj,F)
{
  logLambda = logLambda(MCobj$setup$seasonpar, 2*MCobj$setup$m)
  values = exp(MCobj$mc + logLambda)-F
  mean(values)
}

removeAR = function(X, arobj)
{
  n=length(X)
  result = c()
  for(i in (arobj$order+1):n)
  {
    w = X[i]
    for(j in 1:arobj$order)
    {
      w = w-arobj$ar[j]*X[i-j]
    }
    result = c(result,w)
  }
  result
}

# Calculate density of Up-and-out barrier
# process, by Path Integrals propagated
# with Simpson's Rule
#
# Model: Deterministic + AR(1)
#
# Input
# m         - number of days until expiry
# data      - the data from which we estimate model params
# B         - the upper barrier
# N         - number of points in density representation grid
# K         - number if points in integration grid
# setup     - obj returned from setupPI-function
# zero      - value at which the propagator function is considered zero
# bgwidth   - width of barrier grid, as a fraction of gridspacing
# bgpoints  - number of points in the grid around barrier
# barrierGrid - boolean, true if we should use fine grid around barrier
#
UAObarrierDens = function(m, data, B, N=100, K=1000, setup=NULL, zero=10^-6, bgwidth=5, bgpoints=20, barrierGrid=TRUE)
{
  ptm <- proc.time()

  # check for input errors
  if(tail(data,1)>B)
    stop("Value out of range: B must be larger than S_0")

  # retrieve regular PI setup
  if(is.null(setup))
    setup = setupPI(m,data)
  
  # get seasonality function for t=1:m
  logLambda = logLambda(setup$seasonpar, (m+1):(2*m))
  # calculate log barrier and x-barrier
  logB = log(B)
  XB   = logB-logLambda
  #print(XB)

  # setup grid for h_m(x)
  # let gridmax be slightly above largest barrier
  gridmin = setup$gridrange[1]
  gridspace = (max(XB)-gridmin)/(N-2)
  gridmax = max(XB)+gridspace
  # expand grid up to maximum barrier position
  grid = seq(gridmin,gridmax,length.out=N)

  # calculate and store values of propagator: p(x|x')
  # xMat contains x' values
  # propMat contains corresponding p(x|x')
  # x is given by grid[row index]
  propMat = c()
  tempMat = c()
  xMat = c()
  # test run to find location of probability mass
  for(i in 1:N)
  {
    Xval = grid[i]   # value of x for this col
    condpdf = dnig(grid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
    tempMat = cbind(tempMat, condpdf)
  }
  # find integration grid and propagator
  # densities for each x in grid
  for(i in 1:N)
  {
    # find limits and grid for this propagator row
    # to avoid calculations where propagator is zero
    limits = range(grid[which(tempMat[i,]>zero)])
    #print(limits[2]-limits[1])
    xMat = rbind(xMat,seq(limits[1],limits[2],length.out=K))

    # find densities for this propagator row
    row = c()
    for(j in 1:K)
    {
      row = c(row, dnig(grid[i], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*xMat[i,j])))
    }
    propMat = rbind(propMat,row)
  }

  # find distribution at t=1 (do calculations in deseasonalized space)
  X0 = log(tail(data,1))-logLambda(setup$seasonpar, m)
  pdf = c(0,dnig(grid[2:(N-1)], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4] + setup$arpar*X0)),0)
  spline = splinefun(grid, pdf,method="natural")
  # find H-density function
  hj = knockOutHdens(spline,setup$gridrange,XB[1])
  
  print("starting iteration forward in time")
  # iterate through time steps
  for(t in 2:m)
  {
    # iterate through grid points
    newpdf = rep(0,N)
    for(i in 2:(N-1))
    {
      # perform integration
      dens = 0
      h = diff(xMat[i,1:2])
      integrand = propMat[i,] * hj(xMat[i,])
      for(j in seq(1,K-2,2))
      {
        simpson = (h/3) * (integrand[j] + 4*integrand[j+1] + integrand[j+2])
        dens = dens + simpson
      }
      newpdf[i] = dens
    }


    #--------------------------------------------------------------------------------------------------------
    # Part where we find a fine grid around barrier and calculate pdf in those points
    #--------------------------------------------------------------------------------------------------------

    if(barrierGrid)
    {
      # Find pdf in the close vincinity of the barrier
      # barrier grid
      bgrid = seq(XB[t-1]*setup$arpar-0.5*gridspace*bgwidth, XB[t-1]*setup$arpar+0.5*gridspace*bgwidth, length.out=bgpoints)  

      # calculate and store values of propagator: p(x|x')
      # bxMat contains x' values
      # bpropMat contains corresponding p(x|x')
      # x is given by grid[row index]
      bpropMat = c()
      btempMat = c()
      bxMat = c()

      # test run to find location of probability mass
      for(i in 1:N)
      {
        Xval = grid[i]   # value of x' for this col
        condpdf = dnig(bgrid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
        btempMat = cbind(btempMat, condpdf)
      }
      # find grid and propagator densities for each x
      for(i in 1:length(bgrid))
      {
        # find limits and grid for this propagator row
        # to avoid calculations where propagator is zero
        limits = range(grid[which(btempMat[i,]>zero)])
        bxMat = rbind(bxMat,seq(limits[1],limits[2],length.out=K))

        # find densities for this propagator row
        row = c()
        for(j in 1:K)
        {
          row = c(row, dnig(bgrid[i], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*bxMat[i,j])))
        }
        bpropMat = rbind(bpropMat,row)
      }

      # iterate through grid points
      bpdf = rep(NA,bgpoints)
      for(i in 1:bgpoints)
      {
        # perform integration
        dens = 0
        h = diff(bxMat[i,1:2])
        integrand = bpropMat[i,] * hj(bxMat[i,])
        for(j in seq(1,K-2,2))
        {
          simpson = (h/3) * (integrand[j] + 4*integrand[j+1] + integrand[j+2])
          dens = dens + simpson
        }
        bpdf[i] = dens
      }
      spline = splinefun(c(grid,bgrid),c(newpdf,bpdf),method="natural")
      
    }
    else
      spline = splinefun(grid,newpdf,method="natural")
    
    #--------------------------------------------------------------------------------------------------------


    #find H-density function; h_j
    hj = knockOutHdens(spline,setup$gridrange,XB[t])

    #  if(t%%5==0)
    #  {
    #    xlim = c(0.4,0.7)
    #    sq = seq(xlim[1],xlim[2],0.001)
    #    plot(sq,hj(sq),type="l",xlim=xlim,ylim=c(0,0.001))
    #    points(grid,hj(grid))
    #    points(bgrid,bpdf,pch="x")
    #    stop()
    #  }

    if(t%%30==0)
    {
      print(paste("Progress:",t))
      timediff = proc.time()-ptm
      timeleft = timediff[3]*((m-t)/t)/60
      print(paste("Estimated time left:", timeleft, "minutes."))
    }
  }


  #bspline = splinefun(bgrid,bpdf, method="natural")
  #dev.new()
  #s=seq(XB[t]-0.02,XB[t]+0.02,0.0001)
  #plot(s,gspline(s),type="l")
  #lines(s,hj(s),col=2)
  #lines(s,gspline(s),col=4)
  #lines(bgrid,bpdf,col=3,pch="x")
  #points(grid,newpdf)

  # construct and return result obj
  result=NULL
  result$time = proc.time()-ptm
  if(barrierGrid)
  {
    result$bgrid = bgrid
    result$bpdf = bpdf
  }
  result$newpdf = newpdf
  result$B = B
  result$grid = grid
  result$spline = spline
  result$density = hj
  result$setup = setup
  result
}



# Return knock-out H-density function
#
# Input:
# dens      - density function
# limits    - interval where all mass is located
# B         - the barrier (in appropriate scale)
#
knockOutHdens = function(dens,limits,B)
{
  (function(x){
    val = rep(0,length(x))
    notzero = which(x>limits[1] & x<=B)
    val[notzero] = dens(x[notzero])
    val
  })
}

# Return density function that is
# zero below the barrier
#
# Input:
# dens      - density function
# limits    - interval where all mass is located
# B         - the barrier (in appropriate scale)
#
barOverflow = function(dens,limits,B)
{
  (function(x){
    val = rep(0,length(x))
    notzero = which(x<limits[2] & x>=B)
    val[notzero] = dens(x[notzero])
    val
  })
}


#  Estimate values of up-and-out barrier
#  call options by repeated MC runs
#
#  Input:
#  setup    -  setup object from setupPI()
#  E        -  exercise price of option
#  B        -  barrier value
#  n        -  number of MC runs
#
mcUAObarrier = function(setup, E, B, n=10^6)
{
  ptm <- proc.time()

  # get seasonality function for t=1:m
  logLambda = logLambda(setup$seasonpar, (setup$m+1):(2*setup$m))
  # calculate deseasonalized x-barrier
  Xbar   = log(B)-logLambda

  # create vector for X-values
  Xvals = rep(NA,n)
  #breakstep = c()

  for(i in 1:n)
  {
    # simulate m NIG-residuals
    nigvar = rnig(setup$m,setup$nigpar[1],setup$nigpar[2],setup$nigpar[3],setup$nigpar[4])

    # simulate Autoregressive process X
    X = setup$X0
    alpha = setup$arpar-1
    brokebarrier = FALSE
    for(j in 1:setup$m)
    {
      dX = alpha*X + nigvar[j]
      X = X+dX
      if(X>Xbar[j])
      {
        brokebarrier = TRUE
        #breakstep = c(breakstep,j)
        break
      }
    }
    if(brokebarrier==FALSE)
    {
      Xvals[i] = X
    }


    if((i%%10000)==0)
    {
      print(paste("Progress:",i))
      timeleft = (proc.time()[3]-ptm[3])*(n-i) / (i*60)
      print(paste("Estimated time left:",timeleft,"minutes"))
    }
  }

  # option values
  values = exp(Xvals+tail(logLambda,1))-E
  call = values[which(values>0)]
  put = -values[which(values<0)]
  
  result = NULL
  result$time = proc.time()-ptm
  result$n = n
  result$setup = setup
  result$lambdatail = tail(logLambda,1)
  result$forward = values
  result$call = call
  result$put = put
  result$callval = sum(call)/n
  result$putval = sum(put)/n
  result$Xvals = Xvals
  result$B = B
  result$E = E
  #result$breakstep = breakstep
  result
}


#  Estimate values of up-and-in barrier
#  call options by repeated MC runs
#
#  Input:
#  setup    -  setup object from setupPI()
#  E        -  exercise price of option
#  B        -  barrier value
#  n        -  number of MC runs
#
mcUAIbarrier = function(setup, E, B, n=10^6)
{
  ptm <- proc.time()

  # get seasonality function for t=1:m
  logLambda = logLambda(setup$seasonpar, (setup$m+1):(2*setup$m))
  # calculate deseasonalized x-barrier
  Xbar   = log(B)-logLambda

  # create  for X-values
  Xvals = rep(NA,n)
  #breakstep = c()

  for(i in 1:n)
  {
    # simulate m NIG-residuals
    nigvar = rnig(setup$m,setup$nigpar[1],setup$nigpar[2],setup$nigpar[3],setup$nigpar[4])

    # simulate Autoregressive process X
    X = setup$X0
    alpha = setup$arpar-1
    brokebarrier = FALSE
    for(j in 1:setup$m)
    {
      dX = alpha*X + nigvar[j]
      X = X+dX
      if(X>Xbar[j])
        brokebarrier = TRUE
    }
    if(brokebarrier)
    {
      Xvals[i] = X
    }


    if((i%%10000)==0)
    {
      print(paste("Progress:",i))
      timeleft = (proc.time()[3]-ptm[3])*(n-i) / (i*60)
      print(paste("Estimated time left:",timeleft,"minutes"))
      #print(paste("Estimate proportion:", length(which(!is.na(Xvals)))/i))
    }
  }

  # option values
  values = exp(Xvals+tail(logLambda,1))-E
  call = values[which(values>0)]
  put = -values[which(values<0)]

  
  result = NULL
  result$time = proc.time()-ptm
  result$n = n
  result$setup = setup
  result$call = call
  result$put = put
  result$callval = sum(call)/n
  result$putval = sum(put)/n
  result$Xvals = Xvals
  result$B = B
  result$E = E
  #result$breakstep = breakstep
  result
}



# Calculate density of Up-and-in barrier
# process, by Path Integrals propagated
# with Simpson's Rule
#
# Model: Deterministic + AR(1)
#
# Input
# m         - number of days until expiry
# data      - the data from which we estimate model params
# B         - the upper barrier
# N         - number of points in density representation grid
# K         - number if points in integration grid
# setup     - obj returned from setupPI-function
# zero      - value at which the propagator function is considered zero
# bgwidth   - width of barrier grid, as a fraction of gridspacing
# bgpoints  - number of points in the grid around barrier
# barrierGrid - boolean, true if we should use fine grid around barrier
#
UAIbarrierDens = function(m, data, B, N=100, K=1000, setup=NULL, zero=10^-6, bgwidth=5, bgpoints=20, barrierGrid=TRUE)
{
  ptm <- proc.time()

  # check for input errors
  if(tail(data,1)>B)
    stop("Value out of range: B must be larger than S_0")

  # retrieve regular PI setup
  if(is.null(setup))
    setup = setupPI(m,data)
  
  # get seasonality function for t=1:m
  logLambda = logLambda(setup$seasonpar, (m+1):(2*m))
  # calculate log barrier and x-barrier
  logB = log(B)
  XB   = logB-logLambda
  #print(XB)

  # setup grid for h_m(x) and g_m(x)
  gridmin = setup$gridrange[1]
  gridmax = setup$gridrange[2]
  gridspace = (gridmax-gridmin)/(N-1)
  # expand grid
  grid = seq(gridmin,gridmax,length.out=N)

  # calculate and store values of propagator: p(x|x')
  # xMat contains x' values
  # propMat contains corresponding p(x|x')
  # x is given by grid[row index]
  propMat = c()
  tempMat = c()
  xMat = c()
  # test run to find location of probability mass
  for(i in 1:N)
  {
    Xval = grid[i]   # value of x for this col
    condpdf = dnig(grid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
    tempMat = cbind(tempMat, condpdf)
  }
  # find integration grid and propagator
  # densities for each x in grid
  for(i in 1:N)
  {
    # find limits and grid for this propagator row
    # to avoid calculations where propagator is zero
    limits = range(grid[which(tempMat[i,]>zero)])
    #print(limits[2]-limits[1])
    xMat = rbind(xMat,seq(limits[1],limits[2],length.out=K))

    # find densities for this propagator row
    row = c()
    for(j in 1:K)
    {
      row = c(row, dnig(grid[i], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*xMat[i,j])))
    }
    propMat = rbind(propMat,row)
  }

  # find distribution at t=1 (do calculations in deseasonalized space)
  X0 = log(tail(data,1))-logLambda(setup$seasonpar, m)
  pdf = c(0,dnig(grid[2:(N-1)], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4] + setup$arpar*X0)),0)
  pdfspline = splinefun(grid, pdf,method="natural")
  # find G-density function (=knock-out H-density)
  gj = knockOutHdens(pdfspline,setup$gridrange,XB[1])  
  # find H-density function
  hj = barOverflow(pdfspline,setup$gridrange,XB[1])

  # some test plotting
  #s = seq(min(XB[1:10])-0.1,max(XB[1:10])+0.1,0.0001)
  #plot(s,hj(s),type="l",ylim=c(0,15*max(hj(s))),main="Knock-in density evolving with time", ylab=expression(h[t](x)), xlab="x")
  
  print("starting iteration forward in time")
  # iterate through time steps
  for(t in 2:m)
  {
    # iterate through grid points
    h_pdf = c(0)
    g_pdf = c(0)

    for(i in 2:(N-1))
    {
      # perform integration
      dens = c(0,0)  # = c(h,g)
      # distance between integration grid points
      h = diff(xMat[i,1:2])

      h_integrand = propMat[i,] * hj(xMat[i,])
      g_integrand = propMat[i,] * gj(xMat[i,])
      for(j in seq(1,K-2,2))
      {
        simpson = (h/3) * (h_integrand[j] + 4*h_integrand[j+1] + h_integrand[j+2])
        dens[1] = dens[1] + simpson
        simpson = (h/3) * (g_integrand[j] + 4*g_integrand[j+1] + g_integrand[j+2])
        dens[2] = dens[2] + simpson        
      }
      h_pdf = c(h_pdf,dens[1])
      g_pdf = c(g_pdf,dens[2])
    }
    h_pdf = c(h_pdf,0)
    g_pdf = c(g_pdf,0)



    #--------------------------------------------------------------------------------------------------------
    # Part where we find a fine grid around barrier and calculate pdf in those points
    #--------------------------------------------------------------------------------------------------------

    if(barrierGrid)
    {
      # Find pdf in the close vincinity of the barrier
      # barrier grid
      bpdf = c()
      bgrid = seq(XB[t-1]*setup$arpar-0.5*gridspace*bgwidth, XB[t-1]*setup$arpar+0.5*gridspace*bgwidth, length.out=bgpoints)  

      # calculate and store values of propagator: p(x|x')
      # bxMat contains x' values
      # bpropMat contains corresponding p(x|x')
      # x is given by grid[row index]
      bpropMat = c()
      btempMat = c()
      bxMat = c()

      # test run to find location of probability mass
      for(i in 1:N)
      {
        Xval = grid[i]   # value of x' for this col
        condpdf = dnig(bgrid, setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*Xval))
        btempMat = cbind(btempMat, condpdf)
      }
      # find grid and propagator densities for each x
      for(i in 1:length(bgrid))
      {
        # find limits and grid for this propagator row
        # to avoid calculations where propagator is zero
        limits = range(grid[which(btempMat[i,]>zero)])
        bxMat = rbind(bxMat,seq(limits[1],limits[2],length.out=K))

        # find densities for this propagator row
        row = c()
        for(j in 1:K)
        {
          row = c(row, dnig(bgrid[i], setup$nigpar[1], setup$nigpar[2], setup$nigpar[3], (setup$nigpar[4]+setup$arpar*bxMat[i,j])))
        }
        bpropMat = rbind(bpropMat,row)
      }

      # iterate through grid points
      h_bpdf = c()
      g_bpdf = c()
      for(i in 1:length(bgrid))
      {
        # perform integration
        dens = c(0,0) # c(h,g)
        # distance between integration grid points
        h = diff(bxMat[i,1:2])

        h_integrand = bpropMat[i,] * hj(bxMat[i,])
        g_integrand = bpropMat[i,] * gj(bxMat[i,])

        for(j in seq(1,K-2,2))
        {
          simpson = (h/3) * (h_integrand[j] + 4*h_integrand[j+1] + h_integrand[j+2])
          dens[1] = dens[1] + simpson
          simpson = (h/3) * (g_integrand[j] + 4*g_integrand[j+1] + g_integrand[j+2])
          dens[2] = dens[2] + simpson
        }
        h_bpdf = c(h_bpdf,dens[1])
        g_bpdf = c(g_bpdf,dens[2])
      }
      h_spline = splinefun(c(grid,bgrid),c(h_pdf,h_bpdf),method="natural")
      g_spline = splinefun(c(grid,bgrid),c(g_pdf,g_bpdf),method="natural")
    }
    else
    {
      h_spline = splinefun(grid,h_pdf,method="natural")
      g_spline = splinefun(grid,g_pdf,method="natural")
    }
    
    #--------------------------------------------------------------------------------------------------------


    #find H-density function
    hj =  (function(x){
            val = rep(0,length(x))
            aboveBar = which(x<gridmax & x>XB[t])
            val[aboveBar] = h_spline(x[aboveBar]) + g_spline(x[aboveBar])
            belowBar = which(x>gridmin & x<XB[t])
            val[belowBar] = h_spline(x[belowBar])
            val
          })
    #find G-density function
    gj = knockOutHdens(g_spline,setup$gridrange,XB[t])


    if(t%%30==0)
    {
      print(paste("Progress:",t))
      timediff = proc.time()-ptm
      timeleft = timediff[3]*((m-t)/t)/60
      print(paste("Estimated time left:", timeleft, "minutes."))
    }
  }

  totalmass = integrate(gj,gridmin,XB[t])$val + integrate(hj,gridmin,XB[t])$val +
              integrate(hj,XB[t],gridmax)$val

  #bspline = splinefun(bgrid,bpdf, method="natural")
  #dev.new()
  #s=seq(XB[t]-0.02,XB[t]+0.02,0.0001)
  #plot(s,gspline(s),type="l")
  #lines(s,hj(s),col=2)
  #lines(s,gspline(s),col=4)
  #lines(bgrid,bpdf,col=3,pch="x")
  #points(grid,newpdf)

  # construct and return result obj
  result=NULL
  result$time = proc.time()-ptm
  if(barrierGrid)
  {
    result$bgrid = bgrid
    result$g_bpdf = g_bpdf
    result$h_bpdf = h_bpdf
  }
  result$g_pdf = g_pdf
  result$h_pdf = h_pdf
  result$grid = grid
  result$g_spline = g_spline
  result$h_spline = h_spline
  result$totalmass = totalmass
  # since we have calculated the full probability density
  # i.e. hj + gj
  # we can correct the densities such that p(Omega)=1
  result$density = (function(x){hj(x)/totalmass})
  result$gdensity = (function(x){gj(x)/totalmass})
  result$setup = setup
  result
}


logLambda = function(param,t)
{
  b0 = param[1]
  b1 = param[2]
  b2 = param[3]
  b3 = param[4]
  tau1 = param[5]
  tau2 = param[6]

  result = c()
  for (i in 1:length(t)){
    result = c(result, b0 + b1*cos((tau1 + 2*pi*t[i])/260) + b2*cos((tau2+2*pi*t[i])/5) + b3*t[i])
  }
  result
}

# sum of squares of residuals
SSR = function(params, data)
{
  rsq = data - g(params,1:length(data))
  result <- sum(rsq^2)
  attr(result, "gradient") <- SSRgrad(params,data)
  result
}

SSRgrad = function(params,data)
{
  grad = rep(0,6)
  n = length(data)
  j=1:n

  # translate parameters for readability
  b0 = params[1]
  b1 = params[2]
  b2 = params[3]
  b3 = params[4]
  tau1 = params[5]
  tau2 = params[6]

  grad[1] = 2*sum(-data+b0+b1*cos((tau1+2*pi*j)/260)+b2*cos((tau2+2*pi*j)/5)+b3*j)
  grad[2] = 2*sum(cos((tau1+2*pi*j)/260) * (-data+b1*cos((tau1+2*pi*j)/260)+b0+b3*j+b2*cos((tau2+2*pi*j)/5)))
  grad[3] = 2*sum(cos((tau2+2*pi*j)/5) * (-data+b1*cos((tau1+2*pi*j)/260)+b0+b3*j+b2*cos((tau2+2*pi*j)/5)))
  grad[4] = 2*sum( j*(-data + b3*j + b0 + b1*cos((tau1+2*pi*j)/260) + b2*cos((tau2+2*pi*j)/5)))
  grad[5] = sum(b1* (sin((tau1+2*pi*j)/260)/130) * (data - b1*cos((tau1+2*pi*j)/260) - b0 - b3*j - b2*cos((tau2+2*pi*j)/5)))
  grad[6] = sum(b2* (sin((tau2+2*pi*j)/5)/2.5) * (data - b2*cos((tau2+2*pi*j)/5) - b0 - b3*j - b1*cos((tau1+2*pi*j)/260)))

  grad
}

# ordinary LS method
ordinaryLS = function(data,init=c(0,0,0,0,0,0),typsize=c(5,1,0.1,0.1,0,0), iterlim=250)
{
  result = nlm(SSR,p=init,data=data,typsize=typsize,print.level=1, iterlim=iterlim)
  result
}

nigalpha = function(m2,g1,g2)
{
  sqrt(3*g2 - 4*g1^2) / (sqrt(m2)*(g2-(5/3)*g1^2))
}

nigbeta = function(m2,g1,g2)
{
  g1 / (sqrt(m2)*(g2-(5/3)*g1^2))
}

nigdelta = function(m2,g1,g2)
{
  sqrt(m2*(3*g2-5*g1^2)) / (g2-(4/3)*g1^2)
}

nigmu = function(m1,m2,g1,g2)
{
  m1 - g1*3*sqrt(m2)/(3*g2-4*g1^2)
}

estimateNIGparams = function(X)
{
  m1 = mean(X)
  m2 = var(X)
  g1 = sampleskew(X)
  g2 = samplekurtosis(X)
  
  c(nigalpha(m2,g1,g2),nigbeta(m2,g1,g2),nigdelta(m2,g1,g2),nigmu(m1,m2,g1,g2))
}

samplekurtosis = function(X)
{
  n = length(X)
  diff = X-mean(X)
  k1 = ((1/n)*sum(diff^4)) / ((1/n)*sum(diff^2))^2

  k0 = ((n-1)/((n-2)*(n-3))) * ((n+1)*k1 - 3*(n-1))
  k0
}

sampleskew = function(X)
{
  n = length(X)
  diff = X-mean(X)

  s1 = ((1/n)*sum(diff^3)) / ((1/n)*sum(diff^2))^(3/2)

  s0 = (sqrt(n*(n-1))*s1) / (n-2)
  s0
}

samplemoments = function(X)
{
 c(mean(X),var(X),sampleskew(X),samplekurtosis(X))
}

